/*
 * iwconnect/init.c
 * $Rev: 1739 $
 *
 * Copyright (c) 2008-2010, Reto Gantenbein <reto.gantenbein@gmail.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/**
 * @file init.c
 * @brief Implements functions for iwconnect initialization
 */

/* INCLUDES *******************************************************************/

#include <arpa/inet.h>              // inet_ntoa()
#include <errno.h>                  // errno
#include <fcntl.h>                  // open()
#include <netdb.h>                  // getaddrinfo(), struct addrinfo
#include <signal.h>                 // kill()
#include <stdio.h>                  // sprintf()
#include <stdlib.h>                 // rand()
#include <time.h>                   // time()
#include <unistd.h>                 // gethostname(), getpid()
#include <vif.h>                    // vif_get_names()
#include "iwconnect.h"              // iwc_host_info
#include "iwc_print.h"              // iwc_debug(), iwc_print()
#include "message.h"                // WLAN_PREAMBLE_SIZE
#include "netdevice.h"              // get_mac_by_dev(), get_ip_by_dev()


/* GLOBAL VARIABLES ***********************************************************/

extern char *exec_name;
extern iwc_client_info client_info;
extern iwc_model_info model_info;
extern int num_vif;
extern char vif_names[MAX_NUM_VIF][IFNAMSIZ + 1];
extern int tap_sockets[MAX_NUM_VIF];
extern int net_socket;
extern int ip_family;


/* DEBUG OUTPUT ***************************************************************/

/**
 * Print addr_info structure (for debugging only)
 */
void
iwc_print_addrinfo(struct addrinfo *addr)
{
    char str[INET6_ADDRSTRLEN];
    uint16_t port;
    memset(&str, 0, INET6_ADDRSTRLEN);

    iwc_debug("[AI_FLAGS] AI_PASSIVE = %li, AI_CANONNAME = %li, AI_NUMERICHOST = %li\n",
            AI_PASSIVE, AI_CANONNAME, AI_NUMERICHOST);
    iwc_debug("[AF_FAMILY] AF_UNSPEC = %li, AF_INET = %li, AF_INET6 = %li\n",
            AF_UNSPEC, AF_INET, AF_INET6);
    iwc_debug("[SOCKETS] SOCK_STREAM = %li, SOCK_DGRAM = %li\n",
            SOCK_STREAM, SOCK_DGRAM);
    iwc_debug("[PROTOCOL] IPPROTO_TCP = %li, IPPROTO_UDP = %li\n",
            IPPROTO_TCP, IPPROTO_UDP);
    iwc_debug("ai_flags = %li\n", addr->ai_flags);
    iwc_debug("ai_family = %li\n", addr->ai_family);
    iwc_debug("ai_socktype = %li\n", addr->ai_socktype);
    iwc_debug("ai_protocol = %li\n", addr->ai_protocol);
    iwc_debug("ai_addrlen = %i\n", addr->ai_addrlen);
    if (addr->ai_family == AF_INET)
    {
        inet_ntop(addr->ai_family,
                &((struct sockaddr_in *) addr->ai_addr)->sin_addr,
                str, addr->ai_addrlen);
        port = ntohs(((struct sockaddr_in *) addr->ai_addr)->sin_port);
    }
    if (addr->ai_family == AF_INET6)
    {
        inet_ntop(addr->ai_family,
                &((struct sockaddr_in6 *) addr->ai_addr)->sin6_addr,
                str, addr->ai_addrlen);
        port = ntohs(((struct sockaddr_in6 *) addr->ai_addr)->sin6_port);
    }
    iwc_debug("ai_addr = %p (addr = %s, port = %i)\n",
            addr->ai_addr, str, port);
    iwc_debug("ai_canonname = %p (%s)\n", addr->ai_canonname, addr->ai_canonname);
    iwc_debug("ai_next = %p\n", addr->ai_next);
}

/**
 * Print information about wireless model to debug output
 */
void
iwc_print_model_info(iwc_model_info *minfo)
{
    char addr[INET6_ADDRSTRLEN];
    iwc_host_info *info = &minfo->host_info;
    iwc_debug(",-- Model Information -----------------\n");
    iwc_debug("| Name:\t\t\t%s\n", info->hostname);
    if (ip_family == AF_INET)
    {
        iwc_debug("| IPv4 Address:\t\t%s\n",
                inet_ntoa(((struct sockaddr_in *) &info->addr)->sin_addr));
        iwc_debug("| Port:\t\t\t%u\n",
                ntohs(((struct sockaddr_in *) &info->addr)->sin_port));
    }
    if (ip_family == AF_INET6)
    {
        iwc_debug("| IPv6 Address:\t\t%s\n",
                inet_ntop(info->addr_info.ai_family,
                    &((struct sockaddr_in6 *) &info->addr)->sin6_addr,
                    addr, info->addr_info.ai_addrlen));
        iwc_debug("| Port:\t\t\t%u\n",
                ntohs(((struct sockaddr_in6 *) &info->addr)->sin6_port));
    }
    iwc_debug("`--------------------------------------\n");
}

/**
 * Print information about wireless client to debug output
 */
void
iwc_print_client_info(iwc_client_info *cinfo)
{
    char addr[INET6_ADDRSTRLEN];
    iwc_host_info *info = &cinfo->host_info;
    iwc_debug(",-- Host Information ------------------\n");
    iwc_debug("| Name:\t\t\t%s\n", info->hostname);
    iwc_debug("| Interface:\t\t%s\n", cinfo->host_if);
    iwc_debug("| ID/MAC Address:\t%s\n", cinfo->host_id);
    if (ip_family == AF_INET)
    {
        iwc_debug("| IPv4 Address:\t\t%s\n",
                inet_ntoa(((struct sockaddr_in *) &info->addr)->sin_addr));
        iwc_debug("| Port:\t\t\t%u\n",
                ntohs(((struct sockaddr_in *) &info->addr)->sin_port));
    }
    if (ip_family == AF_INET6)
    {
        iwc_debug("| IPv6 Address:\t\t%s\n",
                inet_ntop(info->addr_info.ai_family,
                        &((struct sockaddr_in6 *) &info->addr)->sin6_addr,
                        addr, info->addr_info.ai_addrlen));
        iwc_debug("| Port:\t\t\t%u\n",
                ntohs(((struct sockaddr_in6 *) &info->addr)->sin6_port));
    }
    iwc_debug("`--------------------------------------\n");
}


/* FUNCTIONS ******************************************************************/

/**
 * Check "IP/hostname:port" pair for validity and fills iwc_model_info
 * @param arg Command line arguments
 */
int
iwc_init_model(const char *model, int port)
{
    struct addrinfo request, *result;
    iwc_host_info *info = &model_info.host_info;
    char cport[6];                               // for port (5 char digits)

    // check hostname length
    if (strlen(model) > HOST_NAME_MAX)
    {
        iwc_error("Destination address to long!\n");
        return -1;
    }

    // get hostname/ip and port tokens
    if (port == -1)
        sprintf(cport, "2424");
    else
        sprintf(cport, "%i", port);

    // prepare connection request
    memset(&request, 0, sizeof(struct addrinfo));
    request.ai_flags = 0;
    request.ai_family = ip_family;          // use the IP protocol we've choosen
    request.ai_socktype = SOCK_DGRAM;       // UDP socket
    request.ai_protocol = 0;                // any protocol
    request.ai_canonname = NULL;            // doesn't matter for getaddrinfo()
    request.ai_addr = NULL;                 // doesn't matter for getaddrinfo()
    request.ai_next = NULL;                 // doesn't matter for getaddrinfo()

    // lookup destination IP (also give destination port and protocol, so we can
    // easily use the returned sockaddr to create our connection) */
    if (getaddrinfo(model, cport, &request, &result) < 0)
    {
        iwc_error("Cannot resolve destination address (%s)\n", gai_strerror(errno));
        return -1;
    }
    while (1)  // iterate through all returned results
    {
        if (ip_family == result->ai_family)
        {
            memmove(&info->addr_info, result, sizeof(struct addrinfo));
            memmove(&info->addr, result->ai_addr, sizeof(struct sockaddr_storage));
            info->addr_info.ai_addr = (struct sockaddr *) &info->addr;
            info->addr_info.ai_canonname = info->hostname;
        }
        if ((result = result->ai_next) == NULL) // check next item in result list
            break;
    }
    freeaddrinfo(result);
    if (info->addr_info.ai_addr == NULL)
    {
        iwc_error("Couldn't find valid address for destination\n");
        return -1;
    }
    // get real hostname
    if (getnameinfo(info->addr_info.ai_addr,
        info->addr_info.ai_addrlen, info->hostname,
        HOST_NAME_MAX, NULL, 0, NI_NAMEREQD) < 0)
            // don't return error, this is not critical!
            iwc_debug("Failed to lookup hostname of destination IP (%s)\n",
                    gai_strerror(errno));

    iwc_print_model_info(&model_info);
    return 0;
}

/**
 *  Initialize client host variables
 */
int
iwc_init_host()
{
    int i = -1;
    char addr_string[INET6_ADDRSTRLEN];
    addr_storage addr;
    struct addrinfo request, *result;
    uint32_t random_int = 0;
    unsigned char id_source[ETH_ALEN];
    iwc_host_info *info = &client_info.host_info;

    // get host name
    if (gethostname(info->hostname, HOST_NAME_MAX) != 0)
    {
        iwc_error("Unable to retrieve hostname (%s)\n", strerror(errno));
        return -1;
    }
    info->hostname[HOST_NAME_MAX] = '\0';   // properly terminate string

    // get IP of physical interface
    memset(&addr, 0, sizeof(addr));
    if (get_ip_by_dev(&addr, client_info.host_if) != 0)
        return -1;

    memset(addr_string, 0, INET6_ADDRSTRLEN);
    inet_ntop(ip_family, &addr, addr_string, INET6_ADDRSTRLEN);

    // prepare connection request
    memset(&request, 0, sizeof(struct addrinfo));
    request.ai_flags = 0;
    request.ai_family = ip_family;          // use the IP protocol we've choosen
    request.ai_socktype = SOCK_DGRAM;       // UDP socket
    request.ai_protocol = 0;                // any protocol
    request.ai_canonname = NULL;            // doesn't matter for getaddrinfo()
    request.ai_addr = NULL;                 // doesn't matter for getaddrinfo()
    request.ai_next = NULL;                 // doesn't matter for getaddrinfo()

    /*  -> if we specify AI_PASSIVE and *node = NULL, ai_addr is 0.0.0.0
            -> not applicable because model needs to know our IP address
        -> if only *node = NULL, ai_addr is localhost (ipv4/ipv6)
        -> if *node = hostname, ai_addr is IP of registered hostname
            -> no choice of interface possible
        -> if *node = ipaddress, ai_addr is IP of needed interface (but only ipv4)
    */
    // Give listen port and protocol, so we can easily use the returned sockaddr
    // to create our connection
    if (getaddrinfo(addr_string, client_info.host_port, &request, &result) < 0)
    {
        iwc_error("Cannot find listen address (%s)\n", gai_strerror(errno));
        iwc_error("%s\n", strerror(errno));
        return -1;
    }
    while (1)   // iterate through all returned results
    {
        if (ip_family == result->ai_family)
        {
            memmove(&info->addr_info, result, sizeof(struct addrinfo));
            memmove(&info->addr, result->ai_addr, sizeof(struct sockaddr_storage));
            info->addr_info.ai_addr = (struct sockaddr *) &info->addr;
            info->addr_info.ai_canonname = info->hostname;
            if (ip_family == AF_INET6)          // set correct interface index
                ((struct sockaddr_in6 *) &info->addr)->sin6_scope_id =
                    if_nametoindex(client_info.host_if);
        }
        if ((result = result->ai_next) == NULL) // check next item in result list
            break;
    }
    freeaddrinfo(result);

    // get physical MAC address
    get_mac_by_dev(client_info.host_mac, client_info.host_if);

    // create host ID from physical MAC address
    if ((client_info.host_mac[0] == 0) && (client_info.host_mac[1] == 0) &&
        (client_info.host_mac[2] == 0) && (client_info.host_mac[3] == 0) &&
        (client_info.host_mac[4] == 0) && (client_info.host_mac[5] == 0)) // we don't have valid mac
    {
        iwc_debug("No valid MAC address found for interface %s\n", client_info.host_if);
        iwc_debug("-> Generating random ID\n");
        srand(time(NULL));
        random_int = random();                  // create random 32bit int...
        id_source[0] = 0x0A;
        id_source[1] = 0xAA;
        id_source[2] = (random_int>>24) & 0xFF; // ... and assign it to last
        id_source[3] = (random_int>>16) & 0xFF; // four bytes of the MAC
        id_source[4] = (random_int>>8) & 0xFF;
        id_source[5] = random_int & 0xFF;
    } else {
        memmove(id_source, client_info.host_mac, ETH_ALEN);
    }
    for (i = 0; i < ETH_ALEN; i++)
        sprintf(client_info.host_id + (2 * i), "%02x", id_source[i]);
    client_info.host_id[2 * ETH_ALEN] = '\0';  // properly terminate string

    iwc_print_client_info(&client_info);
    return 0;
}

/**
 *  Initialize network sockets
 */
int
iwc_init_net()
{
    int i = 0;
    int phy_mtu = 0;
    int vif_mtu = 0;

    // initialize sockets with -1
    net_socket = -1;
    for (i = 0; i < MAX_NUM_VIF; i++)
    {
        tap_sockets[i] = -1;
        *vif_names[i] = '\0';
    }

    // now check virtual wireless interfaces via libvif
    // -> vif_*() functions (see vifd/vif.h)
    if ((num_vif = vif_get_names(vif_names)) < 1)
        return -1;                      // we couldn't find valid interfaces

    iwc_log("Forwarding %i virtual interface(s):\n", num_vif);
    for (i = 0; i < num_vif; i++)
        iwc_log(" %s\n", vif_names[i]);

    // check for proper MTU
    phy_mtu = get_mtu_by_dev(client_info.host_if);
    for (i = 0; i < num_vif; i++)
    {
        vif_mtu = get_mtu_by_dev(vif_names[i]);
        if (phy_mtu < vif_mtu + sizeof(struct ether_header) + DATA_PREAMBLE_SIZE)
        {
            iwc_error("MTU %i on interface '%s' is too small!\n",
                    phy_mtu, client_info.host_if);
            iwc_error("At least %i (%s) + %i required\n", vif_mtu, vif_names[i],
                    sizeof(struct ether_header) + DATA_PREAMBLE_SIZE);
            return -1;
        }
    }

    // create vif sockets
    for (i = 0; i < num_vif; i++)
    {
        if ((tap_sockets[i] = open_tap_socket(vif_names[i])) < 0)
            return -1;
    }

    // initialize listen socket
    if ((net_socket = bind_net_socket(&(client_info.host_info.addr_info))) < 0)
        return -1;

    return 0;
}

/**
 * Create a PID file
 */
int
iwc_init_create_pidfile(const char *path)
{
    char buf[32];

    int fd = open(path, O_WRONLY | O_CREAT, 0640);
    if (fd < 0)
    {
        iwc_error("iwc_init_create_pidfile(): open(%s) failed (%s)",
                path, strerror(errno));
        return 1;
    } else {
        snprintf(buf, sizeof(buf), "%d\n", getpid());
        if (write(fd, buf, strlen(buf)) < 0)
        {
            iwc_error("iwc_init_create_pidfile(): write failed (%s)",
                    path, strerror(errno));
            close(fd);
            return 1;
        }
        close(fd);
    }
    return 0;
}

/**
 *  Remove the PID file
 */
int
iwc_init_remove_pidfile(const char *path)
{
    if (unlink(path) < 0)
    {
        iwc_debug("iwc_model_init_remove_pidfile(): unlink(%s) failed (%s)\n",
                path, strerror(errno));
        return 1;
    }
    return 0;
}

/**
 * Check for valid PID file
 * @return #PID when found, 0 when not found, -1 on error
 */
int
iwc_init_check_pidfile(const char *path)
{
    char buf[32];
    ssize_t rtrn;

    int fd = open(path, O_RDONLY);
    if (fd < 0)
    {
        if (errno == ENOENT)                // no such file or directory
            return 0;

        iwc_error("iwc_init_check_pidfile(): open(%s) failed (%s)",
                path, strerror(errno));
        return -1;
    }
    rtrn = read(fd, buf, sizeof(buf));
    close(fd);

    if (rtrn <= 0)
    {
        if (rtrn == 0)
        {
            iwc_debug("Empty PID file in %s, overriding", path);
            return 0;
        }

        iwc_error("iwc_init_check_pidfile(): read(%s) failed (%s)",
                path, strerror(errno));
        return -1;

    } else {
        // read PID from file
        pid_t pid;

        if (buf[rtrn-1] == '\n')
            rtrn--;
        buf[rtrn] = '\0';
        pid = atoi(buf);
        if (pid == getpid() ||              // check if it is valid process
           (kill(pid, 0) < 0 && errno == ESRCH))
        {
            // doesn't exist
            iwc_debug("iwc_init_check_pidfile(): found PID file (%i) but no such process, ignoring\n",
                pid);
            iwc_init_remove_pidfile(path);
            return 0;
        } else {
            iwc_error("Traffic redirection already running with PID %i\n", pid);
            printf("Notice: Shutdown %s with 'kill -15 %i'\n", exec_name, pid);
            return pid;
        }
    }
}
