/*
 * iwconnect/message.c
 * $Rev: 1432 $
 *
 * Copyright (C) 2008-2010, Reto Gantenbein <reto.gantenbein@gmail.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/**
 * @file message.c
 * @brief Implements message debug functions
 */

/* INCLUDES *******************************************************************/

#include <arpa/inet.h>              // inet_ntop()
#include <netinet/in.h>             // struct in_addr (for wireless.h)
                                    // never ever include <arpa/inet.h> else
                                    // it won't build on Debian and Fedora!
#include <linux/wireless.h>
#include "message.h"
#include "iwc_print.h"


/* FUNCTIONS ******************************************************************/

/** Checks if it is DATA_MSG */
int
is_data_msg(struct msg_header *header)
{
    return header->type == DATA_MSG;
}

/** Checks if it is REGISTRATION_MSG */
int
is_registration_msg(struct msg_header *header)
{
    return header->type == REGISTRATION_MSG;
}

/** Checks if it is DEREGISTRATION_MSG */
int
is_deregistration_msg(struct msg_header *header)
{
    return header->type == DEREGISTRATION_MSG;
}

/** Checks if it is ACK_MSG */
int
is_ack_msg(struct msg_header *header)
{
    return header->type == ACK_MSG;
}

/** Checks if it is CONFIG_MSG */
int
is_config_msg(struct msg_header *header)
{
    return header->type == CONFIG_MSG;
}

/** Print content of WLAN_MSG to debug output */
void
print_data_msg(const char *message)
{
    struct msg_header *header = (struct msg_header *) message;
    struct data_msg *msg = (struct data_msg *) message;
    int frame_size = msg->pkt_length;
    if (frame_size < DATA_PREAMBLE_SIZE)
    {
        iwc_debug("Invalid message size!\n");
        return;
    }
    iwc_debug("[%4i bytes] =========== DATA MESSAGE ================\n",
            frame_size + DATA_PREAMBLE_SIZE);
    iwc_debug("[%4i bytes] ,---------- msg_header ------------------\n",
            sizeof(struct msg_header));
    iwc_debug("(%4i bytes) | type:\t\t%i\n", sizeof(header->type), header->type);
    iwc_debug("(%4i bytes) | id:\t\t%s\n", sizeof(header->id), header->id);
    iwc_debug("(%4i bytes) vif_index:\t\t%i\n", sizeof(uint16_t), msg->vif_index);
    iwc_debug("(%4i bytes) pkg_length:\t\t%i\n", sizeof(uint16_t), frame_size);
    struct ether_header *ether_hdr = (struct ether_header *) msg->wlan_pkt;
    iwc_debug("[%4i bytes] ,---------- raw_packet ------------------\n",
            frame_size);
    iwc_debug("(%4i bytes) | ether_header --------------------------\n",
            sizeof(struct ether_header));
    iwc_debug("(%4i bytes) | | ether_dhost\t%02hhx:%02hhx:%02hhx:%02hhx:%02hhx:%02hhx\n",
            ETH_ALEN, ether_hdr->ether_dhost[0], ether_hdr->ether_dhost[1],
            ether_hdr->ether_dhost[2], ether_hdr->ether_dhost[3],
            ether_hdr->ether_dhost[4], ether_hdr->ether_dhost[5]);
    iwc_debug("(%4i bytes) | | ether_shost\t%02hhx:%02hhx:%02hhx:%02hhx:%02hhx:%02hhx\n",
            ETH_ALEN, ether_hdr->ether_shost[0], ether_hdr->ether_shost[1],
            ether_hdr->ether_shost[2], ether_hdr->ether_shost[3],
            ether_hdr->ether_shost[4], ether_hdr->ether_shost[5]);
    u_int16_t ether_type = ntohs(ether_hdr->ether_type);
    switch (ether_type)
    {
        case ETHERTYPE_IP:
            iwc_debug("(%4i bytes) | | type\t\t%i (ETHERTYPE_IP)\n",
                    sizeof(u_int16_t), ETHERTYPE_IP);
            break;
        case ETHERTYPE_ARP:
            iwc_debug("(%4i bytes) | | type\t\t%i (ETHERTYPE_ARP)\n",
                    sizeof(u_int16_t), ETHERTYPE_ARP);
            break;
        default:
            iwc_debug("(%4i bytes) | | type\t\t%i (unknown)\n",
                    sizeof(u_int16_t), ether_type);
            break;
    }
    if (frame_size > MAX_FRAME_SIZE)
        iwc_debug("====> FRAME SIZE INVALID (too big!) <====\n");
    iwc_debug("------------------------------------------------------\n");
}

/** Print content of CONFIG_MSG to debug output */
void
print_config_msg(const char *message)
{
    struct msg_header *header = (struct msg_header *) message;
    iwc_debug("[%4i bytes] =========== CONFIGURE MESSAGE ===========\n",
            sizeof(struct config_msg));
    iwc_debug("[%4i bytes] ,---------- msg_header ------------------\n",
            sizeof(struct msg_header));
    iwc_debug("(%4i bytes) | type:\t%i\n", sizeof(header->type), header->type);
    iwc_debug("(%4i bytes) | id:\t%s\n", sizeof(header->id), header->id);
    struct config_msg *conf_msg = (struct config_msg *) message;
    iwc_debug("(%4i bytes) vif_index:\t%i\n",
            sizeof(conf_msg->vif_data.vif_index), conf_msg->vif_data.vif_index);
    uint16_t property = conf_msg->vif_data.type;
    union vif_param *value_p = &(conf_msg->vif_data.value);
    switch (property)
    {
        case SIOCSIWFREQ:
            iwc_debug("(%4i bytes) property:\t%4x (SIOCSIWFREQ)\n",
                    sizeof(property), property);
            iwc_debug("(%4i bytes) value:\t%i\n",
                    sizeof(*value_p), value_p->channel);
            break;
        case SIOCSIWTXPOW:
            iwc_debug("(%4i bytes) property:\t%4x (SICSIWTXPOW)\n",
                    sizeof(property), property);
            iwc_debug("(%4i bytes) value:\t%imW\n",
                    sizeof(*value_p), value_p->txpower);
            break;
        case SIOCSIWFRAG:
            iwc_debug("(%4i bytes) property:\t%4x (SICSIWFRAG)\n",
                    sizeof(property), property);
            iwc_debug("(%4i bytes) value:\t%i\n",
                    sizeof(*value_p), value_p->frag);
            break;
        case SIOCSIWRETRY:
            iwc_debug("(%4i bytes) property:\t%4x (SICSIWRETRY)\n",
                    sizeof(property), property);
            iwc_debug("(%4i bytes) value:\t%i\n",
                    sizeof(*value_p), value_p->retry);
            break;
        case SIOCSIWRTS:
            iwc_debug("(%4i bytes) property:\t%4x (SIOCSIWRTS)\n",
                    sizeof(property), property);
            iwc_debug("(%4i bytes) value:\t%iB\n",
                    sizeof(*value_p), value_p->rts);
            break;
        case SIOCSIWSENS:
            iwc_debug("(%4i bytes) property:\t%4x (SICSIWSENS)\n",
                    sizeof(property), property);
            iwc_debug("(%4i bytes) value:\t%imW\n",
                    sizeof(*value_p), value_p->sensitivity);
            break;
        default:
            iwc_debug("(%4i bytes) property:\t%4x (unknown)\n",
                    sizeof(property), property);
            break;
    }
    iwc_debug("------------------------------------------------------\n");
}

/** Print content of REGISTRATION_MSG to debug output */
void
print_reg_msg(const char *message)
{
    int i = 0;

    iwc_debug("[%4i bytes] =========== REGISTRATION MESSAGE ========\n",
            sizeof(struct registration_msg));
    struct msg_header *header = (struct msg_header *) message;
    iwc_debug("[%4i bytes] ,---------- msg_header ------------------\n",
            sizeof(struct msg_header));
    iwc_debug("(%4i bytes) | type:\t%i\n", sizeof(header->type), header->type);
    iwc_debug("(%4i bytes) | id:\t%s\n", sizeof(header->id), header->id);

    struct registration_info *reg_info =
            &((struct registration_msg *) message)->info;
    iwc_debug("[%4i bytes] ,---------- reg_info --------------------\n",
            sizeof(struct registration_info));
    iwc_debug("(%4i bytes) | name:\t%s\n",
            sizeof(reg_info->name), reg_info->name);

    struct sockaddr *addr = (struct sockaddr *) &reg_info->addr;
    iwc_debug("[%4i bytes] | ,---------------- addr ----------------\n",
        sizeof(struct sockaddr_storage));
    iwc_debug("(%4i bytes) | | sin_family:\t%i\n", sizeof(short), addr->sa_family);
    if (addr->sa_family == AF_INET)
    {
        iwc_debug("(%4i bytes) | | sin_port:\t%i\n", sizeof(unsigned short),
            ntohs(((struct sockaddr_in *) addr)->sin_port));
        iwc_debug("(%4i bytes) | | sin_addr:\t%s\n", sizeof(unsigned long),
            inet_ntoa(((struct sockaddr_in *) addr)->sin_addr));
    }
    if (addr->sa_family == AF_INET6)
    {
        char addr_string[INET6_ADDRSTRLEN];
        iwc_debug("(%4i bytes) | | sin6_port:\t%i\n", sizeof(uint16_t),
            ntohs(((struct sockaddr_in6 *) addr)->sin6_port));
        iwc_debug("(%4i bytes) | | sin6_addr:\t%s\n", 16 * sizeof(unsigned char),
            inet_ntop(AF_INET6, &((struct sockaddr_in6 *) addr)->sin6_addr,
                addr_string, INET6_ADDRSTRLEN));
    }
    iwc_debug("(%4i bytes) | msg_id:\t%i\n",
            sizeof(reg_info->msg_id), reg_info->msg_id);
    iwc_debug("(%4i bytes) | num_vif:\t%i\n",
            sizeof(reg_info->num_vif), reg_info->num_vif);

    for (i = 0; i < reg_info->num_vif; i++)
    {
        vif_registration *vif_reg = &reg_info->vif_data[i];
        unsigned char *mac = vif_reg->vif_mac;
        iwc_debug("[%4i bytes] | ,---------------- vif_registration[%i] -\n",
                sizeof(vif_registration), i);
        iwc_debug("(%4i bytes) | | vif_name:\t%s\n",
                IFNAMSIZ + 1, vif_reg->vif_name);
        iwc_debug("(%4i bytes) | | vif_mac:\t%2.2x:%2.2x:%2.2x:%2.2x:%2.2x:%2.2x\n",
                ETH_ALEN, ((unsigned char *) mac)[0], ((unsigned char *) mac)[1],
                ((unsigned char *) mac)[2], ((unsigned char *) mac)[3],
                ((unsigned char *) mac)[4], ((unsigned char *) mac)[5]);
        iwc_debug("(%4i btyes) | | vif_index\t%i\n",
                sizeof(vif_reg->vif_index), vif_reg->vif_index);
        iwc_debug("(%4i bytes) | | txpower:\t%imW\n",
                sizeof(reg_info->vif_data->txpower), vif_reg->txpower);
        iwc_debug("(%4i bytes) | | retry:\t\t%i\n",
                sizeof(reg_info->vif_data->retry), vif_reg->retry);
        iwc_debug("(%4i bytes) | | rts:\t\t%iB\n",
                sizeof(reg_info->vif_data->rts), vif_reg->rts);
        iwc_debug("(%4i bytes) | | fragmentation:\t%iB\n",
                sizeof(reg_info->vif_data->frag), vif_reg->frag);
        iwc_debug("(%4i bytes) | | sensitivity:\t%imW\n",
                sizeof(reg_info->vif_data->sensitivity), vif_reg->sensitivity);
        iwc_debug("(%4i bytes) | | channel:\t%i\n",
                sizeof(reg_info->vif_data->channel), vif_reg->channel);
    }
    iwc_debug("------------------------------------------------------\n");
}

/** Print content of ACK_/DEREGISTRATION_MSG to debug output */
void
print_plain_msg(const char *message)
{
    struct msg_header *header = (struct msg_header *) message;
    if (is_ack_msg(header))
        iwc_debug("[%4i bytes] =========== ACKNOWLEDGE MESSAGE =========\n",
                sizeof(struct plain_msg));
    if (is_deregistration_msg(header))
        iwc_debug("[%4i bytes] =========== DEREGISTRATION MESSAGE ======\n",
                sizeof(struct plain_msg));
    iwc_debug("[%4i bytes] ,---------- msg_header ------------------\n",
            sizeof(struct msg_header));
    iwc_debug("(%4i bytes) | type:\t%i\n", sizeof(header->type), header->type);
    iwc_debug("(%4i bytes) | id:\t%s\n", sizeof(header->id), header->id);
    struct plain_msg *msg = (struct plain_msg *) message;
    iwc_debug("(%4i bytes) msg_id:\t%i\n", sizeof(uint8_t), msg->msg_id);
    iwc_debug("------------------------------------------------------\n");
}

/** Print message content to debug output */
void
debug_msg(const char *message)
{
    struct msg_header *header = (struct msg_header *) message;
    switch (header->type)
    {
        case DATA_MSG:
            print_data_msg(message);
            break;
        case CONFIG_MSG:
            print_config_msg(message);
            break;
        case REGISTRATION_MSG:
            print_reg_msg(message);
            break;
        case ACK_MSG:
            print_plain_msg(message);
            break;
        case DEREGISTRATION_MSG:
            print_plain_msg(message);
            break;
        default:
            iwc_debug("Warning: unknown message\n");
            return;
    }
}
