/*
 * iwconnect/netdevice.c
 * $Rev: 1222 $
 *
 * Copyright (C) 2008-2010, Reto Gantenbein <reto.gantenbein@gmail.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/**
 * @file netdevice.c
 * @brief Implements functions for network device handling
 */

/* INCLUDES *******************************************************************/

#include <arpa/inet.h>              // inet_ntop(), inet_pton()
#include <linux/if.h>               // IFNAMSIZ
#include <linux/if_tun.h>           // IFF_TAP, TUNSETIFF, TUNSETNOCSUM, ...
#include <sys/ioctl.h>              // ioctl()
#include <sys/socket.h>             // socket(), bind(), SIOCGIFHWADDR, ...
#include <errno.h>                  // errno, ENODEV
#include <fcntl.h>                  // open(), O_RDWR
#include <stdio.h>                  // perror()
#include <string.h>                 // memmove(), memset(), strcpy(), ...
#include <unistd.h>                 // close()
#include "netdevice.h"
#include "iwc_print.h"

#define PROC_IFINET6            "/proc/net/if_inet6"
#define IPV6_ADDR_LOOPBACK      0x0010U
#define IPV6_ADDR_LINKLOCAL     0x0020U
#define IPV6_ADDR_SITELOCAL     0x0040U
#define IPV6_ADDR_COMPATv4      0x0080U


/* LOCAL VARIABLES ************************************************************/

extern int ip_family;
int skfd = -1;                      // socket file descriptor to the kernel


/* FUNCTIONS ******************************************************************/

/**
 * Initialize UPD socket file descriptor (IPv4 or IPv6)
 * @return File descriptor, -1 if error
 */
int
init_udp_socket(void)
{
    if ((skfd = socket(ip_family, SOCK_DGRAM, 0)) == -1)
        iwc_error("Failed to open socket (%s)\n", strerror(errno));

    return skfd;
}

/**
 * Return the MAC address of the given interface
 * @param mac 6 bytes char array to fill in the MAC address
 * @param dev_name Device name of requested address
 * @return 0 if success, -1 else
 */
int
get_mac_by_dev(unsigned char *mac, const char *dev_name)
{
    struct ifreq ifr;

    strncpy(ifr.ifr_name, dev_name, IFNAMSIZ);
    if (ioctl(skfd, SIOCGIFHWADDR, &ifr) != 0)
    {
        if (errno == ENODEV)
            iwc_error("No such device (%s)\n", ifr.ifr_name);
        else
            iwc_error("Failed to get MAC address of %s (%s)\n",
                dev_name, strerror(errno));
        return -1;
    }
    memmove(mac, ifr.ifr_hwaddr.sa_data, 6);

    return 0;
}

/**
 * Return IP address of a given interface
 * @param addr Struct to fill in IP address (sockaddr_in or sockaddr_in6)
 * @param dev_name Device name of requested address
 * @return 0 if success, -1 else
 */
int
get_ip_by_dev(addr_storage *addr, const char *dev_name)
{
    struct ifreq ifr = { .ifr_mtu = 0 };
    FILE *f = NULL;
    int plen, scope, dad_status, if_idx;
    static char flags[200];
    char addr6p[8][5];
    char addr6[40], name[IFNAMSIZ];

    switch (ip_family)
    {
        case AF_INET:                                   // we can use ioctl()
            strncpy(ifr.ifr_name, dev_name, IFNAMSIZ);
            if (ioctl(skfd, SIOCGIFADDR, &ifr) < 0)
            {
                if (errno == ENODEV)
                    iwc_error("No such device (%s)\n", ifr.ifr_name);
                else
                    iwc_error("Failed to get IP address of %s (%s)\n",
                            dev_name, strerror(errno));
                return -1;
            }
            // if we don't have addr, we just check for valid device
            if (addr != NULL)
            {
                struct sockaddr_in *sockaddr =
                        (struct sockaddr_in *) &(ifr.ifr_addr);
                memmove(addr, &(*sockaddr).sin_addr, sizeof(struct in_addr));
            }
            break;

        /* IPv6 code partly taken from net-tools:lib/interface.c
           (c) Andi Kleen, 1998 */
        case AF_INET6:                                      // read from proc
            if ((f = fopen(PROC_IFINET6, "r")) != NULL)
            {
                while (fscanf(f, "%4s%4s%4s%4s%4s%4s%4s%4s %02x %02x %02x %02x %20s\n",
                          addr6p[0], addr6p[1], addr6p[2], addr6p[3],
                          addr6p[4], addr6p[5], addr6p[6], addr6p[7],
                      &if_idx, &plen, &scope, &dad_status, name) != EOF)
                {
                    if (!strcmp(dev_name, name))
                    {
                        sprintf(addr6, "%s:%s:%s:%s:%s:%s:%s:%s",
                            addr6p[0], addr6p[1], addr6p[2], addr6p[3],
                            addr6p[4], addr6p[5], addr6p[6], addr6p[7]);
                        if (addr != NULL)
                            inet_pton(AF_INET6, addr6, addr);

                        // TODO: Do we need this?
                        //~ printf("  scopeid 0x%x", scope);
                        flags[0] = '<'; flags[1] = 0;
                        if (scope & IPV6_ADDR_COMPATv4) {
                            strcat(flags, "compat,");
                            scope -= IPV6_ADDR_COMPATv4;
                        }
                        if (scope == 0)
                            strcat(flags, "global,");
                        if (scope & IPV6_ADDR_LINKLOCAL)
                            strcat(flags, "link,");
                        if (scope & IPV6_ADDR_SITELOCAL)
                            strcat(flags, "site,");
                        if (scope & IPV6_ADDR_LOOPBACK)
                            strcat(flags, "host,");
                        if (flags[strlen(flags)-1] == ',')
                            flags[strlen(flags)-1] = '>';
                        else
                            flags[strlen(flags)-1] = 0;
                        //~ printf("%s\n", flags);
                    }
                }
            } else {
                iwc_error("Kernel doesn't support IPv6 (%s): %s\n",
                        PROC_IFINET6, strerror(errno));
                return -1;
            }
            break;
    }

    return 0;
}

/**
 * Return the MTU of the given interfaces
 * @param dev_name Virtual wireless device name
 * @return MTU in Bytes
 */
int
get_mtu_by_dev(const char *dev_name)
{
    struct ifreq ifr;

    strncpy(ifr.ifr_name, dev_name, IFNAMSIZ);
    if (ioctl(skfd, SIOCGIFMTU, &ifr) < 0)
    {
        if (errno == ENODEV)
            iwc_error("No such device (%s)\n", ifr.ifr_name);
        else
            iwc_error("Failed to get MTU size of %s (%s)\n", \
                    dev_name, strerror(errno));
        return -1;
    }
//  iwc_debug("get_mtu_by_dev(): %s %i bytes\n", dev_name, ifr.ifr_mtu);

    return ifr.ifr_mtu;
}

/**
 * Create UDP network socket and bind to given address
 * @param addr Address info structure to bind socket
 * @return Socket file descriptor, -1 if error
 */
int
bind_net_socket(const struct addrinfo *addr)
{
    int fd = -1;

    if ((fd = socket(addr->ai_family, addr->ai_socktype, addr->ai_protocol)) < 0)
    {
        iwc_error("Failed to open socket (%s)\n", strerror(errno));
        return -1;
    }

    if (bind(fd, (struct sockaddr *) addr->ai_addr, addr->ai_addrlen) == -1)
    {
        close(fd);
        iwc_error("Couldn't bind to local interface (%s)\n", strerror(errno));
        return -1;
    }

    return fd;
}

/**
 * Return file descriptor for TAP device
 * @param dev_name Virtual wireless device name
 * @return Character device file descriptor, -1 if error
 */
int
open_tap_socket(const char *dev_name)
{
    int fd = 0, err = 0;
    struct ifreq ifr;

    if ((fd = open("/dev/net/tun", O_RDWR)) < 0)
    {
        iwc_error("Failed to open /dev/net/tun (%s)\n", strerror(errno));
        return -1;
    }
    memset(&ifr, 0, sizeof(ifr));
    ifr.ifr_flags = IFF_TAP | IFF_NO_PI;        // set flag for TAP device
                                                // we don't need packet info
    if (*dev_name)
       memmove(ifr.ifr_name, dev_name, IFNAMSIZ);
    err = ioctl(fd, TUNSETIFF, (void *) &ifr);  // select interface
    err |= ioctl(fd, TUNSETNOCSUM, 1);          // disable checksumming
    if (err != 0)
    {
        close(fd);
        iwc_error("Ioctl() on device %s failed (%s)\n",
                dev_name, strerror(errno));
        return -1;
    }

    return fd;
}
